@@ -1,5 +1,5 @@
 /*********************************************************
- * Copyright (C) 1998-2015 VMware, Inc. All rights reserved.
+ * Copyright (C) 1998-2017 VMware, Inc. All rights reserved.
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU Lesser General Public License as published
@@ -35,6 +35,10 @@
 #endif
 #include "vmware.h"
 
+#if defined(__cplusplus)
+extern "C" {
+#endif
+
 #if defined __APPLE__ && !vm_x86_64
 /*
  * Bug 471584: Mac OS X 10.6's valloc() implementation for 32-bit
@@ -381,4 +385,8 @@ Aligned_Realloc(void *buf,   // IN
 #endif
 }
 
+#if defined(__cplusplus)
+}  // extern "C"
+#endif
+
 #endif
