@@ -1639,15 +1639,22 @@ TOOLS_VERSION_UINT_TO_COMPONENTS(const ToolsVersion toolsVersion,   // IN
 #define   TOOLS_VERSION_BALLPEENHAMMER_PATCH2_V_BASE 6
 
 #ifndef RC_INVOKED
-#define   TOOLS_VERSION_NEXT TOOLS_VERSION_TO_UINT(TOOLS_VERSION_NEXT_V)
+#define   TOOLS_VERSION_CRESCENT_WRENCH_RELEASE TOOLS_VERSION_TO_UINT(TOOLS_VERSION_CRESCENT_WRENCH_RELEASE_V)
 #endif /* RC_INVOKED */
-#define   TOOLS_VERSION_NEXT_V_MJR  11
-#define   TOOLS_VERSION_NEXT_V_MNR  2
-#define   TOOLS_VERSION_NEXT_V_BASE 0
+#define   TOOLS_VERSION_CRESCENT_WRENCH_RELEASE_V_MJR  11
+#define   TOOLS_VERSION_CRESCENT_WRENCH_RELEASE_V_MNR  2
+#define   TOOLS_VERSION_CRESCENT_WRENCH_RELEASE_V_BASE 0
 
-#define   TOOLS_VERSION_CURRENT        TOOLS_VERSION_NEXT
-#define   TOOLS_VERSION_CURRENT_STR    TOOLS_VERSION_TO_STR(TOOLS_VERSION_NEXT)
-#define   TOOLS_VERSION_CURRENT_CSV    TOOLS_VERSION_TO_CSV(TOOLS_VERSION_NEXT)
+#ifndef RC_INVOKED
+#define   TOOLS_VERSION_CRESCENT_WRENCH_UPDATE1 TOOLS_VERSION_TO_UINT(TOOLS_VERSION_CRESCENT_WRENCH_UPDATE1_V)
+#endif /* RC_INVOKED */
+#define   TOOLS_VERSION_CRESCENT_WRENCH_UPDATE1_V_MJR  11
+#define   TOOLS_VERSION_CRESCENT_WRENCH_UPDATE1_V_MNR  2
+#define   TOOLS_VERSION_CRESCENT_WRENCH_UPDATE1_V_BASE 5
+
+#define   TOOLS_VERSION_CURRENT        TOOLS_VERSION_CRESCENT_WRENCH_UPDATE1
+#define   TOOLS_VERSION_CURRENT_STR    TOOLS_VERSION_TO_STR(TOOLS_VERSION_CRESCENT_WRENCH_UPDATE1)
+#define   TOOLS_VERSION_CURRENT_CSV    TOOLS_VERSION_TO_CSV(TOOLS_VERSION_CRESCENT_WRENCH_UPDATE1)
 
 /*
  * The extended Tools version is the current Tools version with the
