@@ -1,5 +1,5 @@
 /*********************************************************
- * Copyright (C) 2009-2015 VMware, Inc. All rights reserved.
+ * Copyright (C) 2009-2019 VMware, Inc. All rights reserved.
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU Lesser General Public License as published
@@ -42,16 +42,20 @@ extern "C" {
 #include "copyPasteCompat.h"
 #include "cpName.h"
 #include "cpNameUtil.h"
-#include "dnd.h"
 #include "dndClipboard.h"
-#include "dndMsg.h"
-#include "file.h"
 #include "hgfsUri.h"
-#include "hostinfo.h"
 #include "rpcout.h"
-#include "vmblock.h"
 }
 
+#ifdef USE_UINPUT
+#include "fakeMouseWayland/fakeMouseWayland.h"
+#endif
+#include "dnd.h"
+#include "dndMsg.h"
+#include "hostinfo.h"
+#include "file.h"
+#include "vmblock.h"
+
 /* IsXExtensionPointer may be not defined with old Xorg. */
 #ifndef IsXExtensionPointer
 #define IsXExtensionPointer 4
@@ -74,11 +78,13 @@ DnDUIX11::DnDUIX11(ToolsAppCtx *ctx)
     : mCtx(ctx),
       mDnD(NULL),
       mDetWnd(NULL),
+      mClipboard(),
       mBlockCtrl(NULL),
       mHGGetFileStatus(DND_FILE_TRANSFER_NOT_STARTED),
       mBlockAdded(false),
       mGHDnDInProgress(false),
       mGHDnDDataReceived(false),
+      mGHDnDDropOccurred(false),
       mUnityMode(false),
       mInHGDrag(false),
       mEffect(DROP_NONE),
@@ -88,7 +94,10 @@ DnDUIX11::DnDUIX11(ToolsAppCtx *ctx)
       mNumPendingRequest(0),
       mDestDropTime(0),
       mTotalFileSize(0),
-      mOrigin(0, 0)
+      mOrigin(0, 0),
+      mUseUInput(false),
+      mScreenWidth(0),
+      mScreenHeight(0)
 {
    TRACE_CALL();
 
@@ -102,6 +111,20 @@ DnDUIX11::DnDUIX11(ToolsAppCtx *ctx)
     * corners for now.
     */
    OnWorkAreaChanged(Gdk::Screen::get_default());
+
+#ifdef USE_UINPUT
+   //Initialize the uinput device if available
+   if (ctx->uinputFD != -1) {
+      Screen * scrn = DefaultScreenOfDisplay(XOpenDisplay(NULL));
+      if (FakeMouse_Init(ctx->uinputFD, scrn->width, scrn->height)) {
+         mUseUInput = true;
+         mScreenWidth = scrn->width;
+         mScreenHeight = scrn->height;
+      }
+   }
+#endif
+
+   g_debug("%s: Use UInput? %d.\n", __FUNCTION__, mUseUInput);
 }
 
 
@@ -128,7 +151,7 @@ DnDUIX11::~DnDUIX11()
        && !mHGStagingDir.empty()) {
       uint64 totalSize = File_GetSizeEx(mHGStagingDir.c_str());
       if (mTotalFileSize != totalSize) {
-         g_debug("%s: deleting %s, expecting %"FMT64"d, finished %"FMT64"d\n",
+         g_debug("%s: deleting %s, expecting %" FMT64 "u, finished %" FMT64 "u\n",
                  __FUNCTION__, mHGStagingDir.c_str(),
                  mTotalFileSize, totalSize);
          DnD_DeleteStagingFiles(mHGStagingDir.c_str(), FALSE);
@@ -201,19 +224,19 @@ DnDUIX11::Init()
    CONNECT_SIGNAL(mDnD, updateUnityDetWndChanged, OnUpdateUnityDetWnd);
 
    /* Set Gtk+ callbacks for source. */
-   CONNECT_SIGNAL(mDetWnd, signal_drag_begin(),        OnGtkDragBegin);
-   CONNECT_SIGNAL(mDetWnd, signal_drag_data_get(),     OnGtkDragDataGet);
-   CONNECT_SIGNAL(mDetWnd, signal_drag_end(),          OnGtkDragEnd);
-   CONNECT_SIGNAL(mDetWnd, signal_enter_notify_event(), GtkEnterEventCB);
-   CONNECT_SIGNAL(mDetWnd, signal_leave_notify_event(), GtkLeaveEventCB);
-   CONNECT_SIGNAL(mDetWnd, signal_map_event(),         GtkMapEventCB);
-   CONNECT_SIGNAL(mDetWnd, signal_unmap_event(),       GtkUnmapEventCB);
-   CONNECT_SIGNAL(mDetWnd, signal_realize(),           GtkRealizeEventCB);
-   CONNECT_SIGNAL(mDetWnd, signal_unrealize(),         GtkUnrealizeEventCB);
-   CONNECT_SIGNAL(mDetWnd, signal_motion_notify_event(), GtkMotionNotifyEventCB);
-   CONNECT_SIGNAL(mDetWnd, signal_configure_event(),   GtkConfigureEventCB);
-   CONNECT_SIGNAL(mDetWnd, signal_button_press_event(), GtkButtonPressEventCB);
-   CONNECT_SIGNAL(mDetWnd, signal_button_release_event(), GtkButtonReleaseEventCB);
+   CONNECT_SIGNAL(mDetWnd->GetWnd(), signal_drag_begin(),        OnGtkDragBegin);
+   CONNECT_SIGNAL(mDetWnd->GetWnd(), signal_drag_data_get(),     OnGtkDragDataGet);
+   CONNECT_SIGNAL(mDetWnd->GetWnd(), signal_drag_end(),          OnGtkDragEnd);
+   CONNECT_SIGNAL(mDetWnd->GetWnd(), signal_enter_notify_event(), GtkEnterEventCB);
+   CONNECT_SIGNAL(mDetWnd->GetWnd(), signal_leave_notify_event(), GtkLeaveEventCB);
+   CONNECT_SIGNAL(mDetWnd->GetWnd(), signal_map_event(),         GtkMapEventCB);
+   CONNECT_SIGNAL(mDetWnd->GetWnd(), signal_unmap_event(),       GtkUnmapEventCB);
+   CONNECT_SIGNAL(mDetWnd->GetWnd(), signal_realize(),           GtkRealizeEventCB);
+   CONNECT_SIGNAL(mDetWnd->GetWnd(), signal_unrealize(),         GtkUnrealizeEventCB);
+   CONNECT_SIGNAL(mDetWnd->GetWnd(), signal_motion_notify_event(), GtkMotionNotifyEventCB);
+   CONNECT_SIGNAL(mDetWnd->GetWnd(), signal_configure_event(),   GtkConfigureEventCB);
+   CONNECT_SIGNAL(mDetWnd->GetWnd(), signal_button_press_event(), GtkButtonPressEventCB);
+   CONNECT_SIGNAL(mDetWnd->GetWnd(), signal_button_release_event(), GtkButtonReleaseEventCB);
 
 #undef CONNECT_SIGNAL
 
@@ -251,7 +274,7 @@ DnDUIX11::InitGtk()
    TRACE_CALL();
 
    /* Construct supported target list for HG DnD. */
-   std::list<Gtk::TargetEntry> targets;
+   std::vector<Gtk::TargetEntry> targets;
 
    /* File DnD. */
    targets.push_back(Gtk::TargetEntry(DRAG_TARGET_NAME_URI_LIST));
@@ -259,6 +282,7 @@ DnDUIX11::InitGtk()
    /* RTF text DnD. */
    targets.push_back(Gtk::TargetEntry(TARGET_NAME_APPLICATION_RTF));
    targets.push_back(Gtk::TargetEntry(TARGET_NAME_TEXT_RICHTEXT));
+   targets.push_back(Gtk::TargetEntry(TARGET_NAME_TEXT_RTF));
 
    /* Plain text DnD. */
    targets.push_back(Gtk::TargetEntry(TARGET_NAME_UTF8_STRING));
@@ -279,16 +303,16 @@ DnDUIX11::InitGtk()
     * but in our case, we will call drag_get_data during DragMotion, and
     * will cause X dead with DEST_DEFAULT_ALL. The reason is unclear.
     */
-   mDetWnd->drag_dest_set(targets, Gtk::DEST_DEFAULT_MOTION,
-                          Gdk::ACTION_COPY | Gdk::ACTION_MOVE);
+   mDetWnd->GetWnd()->drag_dest_set(targets, Gtk::DEST_DEFAULT_MOTION,
+                                    Gdk::ACTION_COPY | Gdk::ACTION_MOVE);
 
-   mDetWnd->signal_drag_leave().connect(
+   mDetWnd->GetWnd()->signal_drag_leave().connect(
       sigc::mem_fun(this, &DnDUIX11::OnGtkDragLeave));
-   mDetWnd->signal_drag_motion().connect(
+   mDetWnd->GetWnd()->signal_drag_motion().connect(
       sigc::mem_fun(this, &DnDUIX11::OnGtkDragMotion));
-   mDetWnd->signal_drag_drop().connect(
+   mDetWnd->GetWnd()->signal_drag_drop().connect(
       sigc::mem_fun(this, &DnDUIX11::OnGtkDragDrop));
-   mDetWnd->signal_drag_data_received().connect(
+   mDetWnd->GetWnd()->signal_drag_data_received().connect(
       sigc::mem_fun(this, &DnDUIX11::OnGtkDragDataReceived));
 }
 
@@ -349,25 +373,50 @@ DnDUIX11::OnSrcDragBegin(const CPClipboard *clip,       // IN
    Glib::RefPtr<Gtk::TargetList> targets;
    Gdk::DragAction actions;
    GdkEventMotion event;
+   int mouseX = mOrigin.get_x() + DRAG_DET_WINDOW_WIDTH / 2;
+   int mouseY = mOrigin.get_y() + DRAG_DET_WINDOW_WIDTH / 2;
 
    TRACE_CALL();
 
    CPClipboard_Clear(&mClipboard);
    CPClipboard_Copy(&mClipboard, clip);
 
+#ifdef USE_UINPUT
+   if (mUseUInput) {
+      /*
+       * Check if the screen size changes, if so then update the
+       * uinput device.
+       */
+      Screen * scrn = DefaultScreenOfDisplay(XOpenDisplay(NULL));
+      if (   (scrn->width != mScreenWidth)
+          || (scrn->height != mScreenHeight)) {
+         g_debug("%s: Update uinput device. prew:%d, preh:%d, w:%d, h:%d\n",
+                 __FUNCTION__,
+                 mScreenWidth,
+                 mScreenHeight,
+                 scrn->width,
+                 scrn->height);
+         mScreenWidth = scrn->width;
+         mScreenHeight = scrn->height;
+         FakeMouse_Update(mScreenWidth, mScreenHeight);
+      }
+   }
+#endif
+
    /*
     * Before the DnD, we should make sure that the mouse is released
     * otherwise it may be another DnD, not ours. Send a release, then
     * a press here to cover this case.
     */
-   SendFakeXEvents(false, true, false, false, false, 0, 0);
-   SendFakeXEvents(true, true, true, true, true, mOrigin.get_x(), mOrigin.get_y());
+
+   SendFakeXEvents(true, true, false, true, true, mouseX, mouseY);
+   SendFakeXEvents(false, true, true, false, true, mouseX, mouseY);
 
    /*
     * Construct the target and action list, as well as a fake motion notify
     * event that's consistent with one that would typically start a drag.
     */
-   targets = Gtk::TargetList::create(std::list<Gtk::TargetEntry>());
+   targets = Gtk::TargetList::create(std::vector<Gtk::TargetEntry>());
 
    if (CPClipboard_ItemExists(&mClipboard, CPFORMAT_FILELIST)) {
       mHGStagingDir = stagingDir;
@@ -400,13 +449,14 @@ DnDUIX11::OnSrcDragBegin(const CPClipboard *clip,       // IN
    if (CPClipboard_ItemExists(&mClipboard, CPFORMAT_RTF)) {
       targets->add(Glib::ustring(TARGET_NAME_APPLICATION_RTF));
       targets->add(Glib::ustring(TARGET_NAME_TEXT_RICHTEXT));
+      targets->add(Glib::ustring(TARGET_NAME_TEXT_RTF));
    }
 
    actions = Gdk::ACTION_COPY | Gdk::ACTION_MOVE;
 
    /* TODO set the x/y coords to the actual drag initialization point. */
    event.type = GDK_MOTION_NOTIFY;
-   event.window = mDetWnd->get_window()->gobj();
+   event.window = mDetWnd->GetWnd()->get_window()->gobj();
    event.send_event = false;
    event.time = GDK_CURRENT_TIME;
    event.x = 10;
@@ -414,12 +464,17 @@ DnDUIX11::OnSrcDragBegin(const CPClipboard *clip,       // IN
    event.axes = NULL;
    event.state = GDK_BUTTON1_MASK;
    event.is_hint = 0;
+#ifndef GTK3
    event.device = gdk_device_get_core_pointer();
+#else
+   GdkDeviceManager* manager = gdk_display_get_device_manager(gdk_window_get_display(event.window));
+   event.device = gdk_device_manager_get_client_pointer(manager);
+#endif
    event.x_root = mOrigin.get_x();
    event.y_root = mOrigin.get_y();
 
    /* Tell Gtk that a drag should be started from this widget. */
-   mDetWnd->drag_begin(targets, actions, 1, (GdkEvent *)&event);
+   mDetWnd->GetWnd()->drag_begin(targets, actions, 1, (GdkEvent *)&event);
    mBlockAdded = false;
    mHGGetFileStatus = DND_FILE_TRANSFER_NOT_STARTED;
    SourceDragStartDone();
@@ -456,9 +511,14 @@ DnDUIX11::OnSrcCancel()
     * flybacks when we cancel as user moves mouse in and out of destination
     * window in a H->G DnD.
     */
-   OnUpdateDetWnd(true, 0, 0);
-   SendFakeXEvents(true, true, false, true, true, mOrigin.get_x(), mOrigin.get_y());
+   OnUpdateDetWnd(true, mOrigin.get_x(), mOrigin.get_y());
+   SendFakeXEvents(true, true, false, true, true,
+                   mOrigin.get_x() + DRAG_DET_WINDOW_WIDTH / 2,
+                   mOrigin.get_y() + DRAG_DET_WINDOW_WIDTH / 2);
    OnUpdateDetWnd(false, 0, 0);
+   SendFakeXEvents(false, false, false, false, true,
+                   mMousePosX,
+                   mMousePosY);
    mInHGDrag = false;
    mHGGetFileStatus = DND_FILE_TRANSFER_NOT_STARTED;
    mEffect = DROP_NONE;
@@ -628,12 +688,12 @@ DnDUIX11::OnUpdateDetWnd(bool show,     // IN: show (true) or hide (false)
 {
    g_debug("%s: enter 0x%lx show %d x %d y %d\n",
          __FUNCTION__,
-         (unsigned long) mDetWnd->get_window()->gobj(), show, x, y);
+         (unsigned long) mDetWnd->GetWnd()->get_window()->gobj(), show, x, y);
 
    /* If the window is being shown, move it to the right place. */
    if (show) {
-      x = MAX(x - DRAG_DET_WINDOW_WIDTH / 2, 0);
-      y = MAX(y - DRAG_DET_WINDOW_WIDTH / 2, 0);
+      x = MAX(x - DRAG_DET_WINDOW_WIDTH / 2, mOrigin.get_x());
+      y = MAX(y - DRAG_DET_WINDOW_WIDTH / 2, mOrigin.get_y());
 
       mDetWnd->Show();
       mDetWnd->Raise();
@@ -678,7 +738,7 @@ DnDUIX11::OnUpdateUnityDetWnd(bool show,         // IN: show (true) or hide (fal
 {
    g_debug("%s: enter 0x%lx unityID 0x%x\n",
          __FUNCTION__,
-         (unsigned long) mDetWnd->get_window()->gobj(),
+         (unsigned long) mDetWnd->GetWnd()->get_window()->gobj(),
          unityWndId);
 
    if (show && ((unityWndId > 0) || bottom)) {
@@ -763,7 +823,11 @@ DnDUIX11::OnMoveMouse(int32 x,  // IN: Pointer x-coord
       // If we are the context of a DnD, send DnD feedback to the source.
 
       DND_DROPEFFECT effect;
+#ifndef GTK3
       effect = ToDropEffect((Gdk::DragAction)(mDragCtx->action));
+#else
+      effect = ToDropEffect((Gdk::DragAction)(gdk_drag_context_get_selected_action(mDragCtx)));
+#endif
       if (effect != mEffect) {
          mEffect = effect;
          g_debug("%s: Updating feedback\n", __FUNCTION__);
@@ -836,7 +900,7 @@ DnDUIX11::OnGtkDragMotion(
    Gdk::DragAction srcActions;
    Gdk::DragAction suggestedAction;
    Gdk::DragAction dndAction = (Gdk::DragAction)0;
-   Glib::ustring target = mDetWnd->drag_dest_find_target(dc);
+   Glib::ustring target = mDetWnd->GetWnd()->drag_dest_find_target(dc);
 
    if (!mDnD->IsDnDAllowed()) {
       g_debug("%s: No dnd allowed!\n", __FUNCTION__);
@@ -851,11 +915,16 @@ DnDUIX11::OnGtkDragMotion(
     *       Gdk::DragContext::get_targets, but API/ABI broke between 2.10 and
     *       2.12, so we work around it like this for now.
     */
+#ifndef GTK3
    Glib::ListHandle<std::string, Gdk::AtomStringTraits> targets(
       dc->gobj()->targets, Glib::OWNERSHIP_NONE);
+#else
+   Glib::ListHandle<std::string, Gdk::AtomStringTraits> targets(
+      gdk_drag_context_list_targets(dc->gobj()), Glib::OWNERSHIP_NONE);
+#endif
 
-   std::list<Glib::ustring> as = targets;
-   std::list<Glib::ustring>::iterator result;
+   std::vector<Glib::ustring> as = targets;
+   std::vector<Glib::ustring>::iterator result;
    char *pid;
    pid = Str_Asprintf(NULL, "guest-dnd-target %d", static_cast<int>(getpid()));
    if (pid) {
@@ -871,7 +940,7 @@ DnDUIX11::OnGtkDragMotion(
 
    mDragCtx = dc->gobj();
 
-   if (target != "") {
+   if (target != Gdk::AtomString::to_cpp_type(GDK_NONE)) {
       /*
        * We give preference to the suggested action from the source, and prefer
        * copy over move.
@@ -1014,7 +1083,6 @@ DnDUIX11::OnGtkDragDataGet(
    guint info,                                  // UNUSED
    guint time)                                  // IN: event timestamp
 {
-   size_t index = 0;
    std::string str;
    std::string uriList;
    std::string stagingDirName;
@@ -1040,6 +1108,7 @@ DnDUIX11::OnGtkDragDataGet(
 
    if (   target == DRAG_TARGET_NAME_URI_LIST
        && CPClipboard_GetItem(&mClipboard, CPFORMAT_FILELIST, &buf, &sz)) {
+      size_t index = 0;
 
       /* Provide path within vmblock file system instead of actual path. */
       stagingDirName = GetLastDirName(mHGStagingDir);
@@ -1125,14 +1194,14 @@ DnDUIX11::OnGtkDragDataGet(
 
    if (   TargetIsPlainText(target)
        && CPClipboard_GetItem(&mClipboard, CPFORMAT_TEXT, &buf, &sz)) {
-      g_debug("%s: providing plain text, size %"FMTSZ"u\n", __FUNCTION__, sz);
+      g_debug("%s: providing plain text, size %" FMTSZ "u\n", __FUNCTION__, sz);
       selection_data.set(target.c_str(), (const char *)buf);
       return;
    }
 
    if (   TargetIsRichText(target)
        && CPClipboard_GetItem(&mClipboard, CPFORMAT_RTF, &buf, &sz)) {
-      g_debug("%s: providing rtf text, size %"FMTSZ"u\n", __FUNCTION__, sz);
+      g_debug("%s: providing rtf text, size %" FMTSZ "u\n", __FUNCTION__, sz);
       selection_data.set(target.c_str(), (const char *)buf);
       return;
    }
@@ -1305,11 +1374,11 @@ DnDUIX11::OnGtkDragDrop(
 
    Glib::ustring target;
 
-   target = mDetWnd->drag_dest_find_target(dc);
+   target = mDetWnd->GetWnd()->drag_dest_find_target(dc);
    g_debug("%s: calling drag_finish\n", __FUNCTION__);
    dc->drag_finish(true, false, time);
 
-   if (target == "") {
+   if (target == Gdk::AtomString::to_cpp_type(GDK_NONE)) {
       g_debug("%s: No valid data on clipboard.\n", __FUNCTION__);
       return false;
    }
@@ -1348,10 +1417,6 @@ DnDUIX11::OnGtkDragDrop(
 bool
 DnDUIX11::SetCPClipboardFromGtk(const Gtk::SelectionData& sd) // IN
 {
-   char *newPath;
-   char *newRelPath;
-   size_t newPathLen;
-   size_t index = 0;
    DnDFileList fileList;
    DynBuf buf;
    uint64 totalSize = 0;
@@ -1366,6 +1431,10 @@ DnDUIX11::SetCPClipboardFromGtk(const Gtk::SelectionData& sd) // IN
        * one for just the last path component.
        */
       utf::string source = sd.get_data_as_string().c_str();
+      size_t index = 0;
+      char *newPath;
+      size_t newPathLen;
+
       g_debug("%s: Got file list: [%s]\n", __FUNCTION__, source.c_str());
 
       if (sd.get_data_as_string().length() == 0) {
@@ -1393,7 +1462,9 @@ DnDUIX11::SetCPClipboardFromGtk(const Gtk::SelectionData& sd) // IN
       while ((newPath = DnD_UriListGetNextFile(source.c_str(),
                                                &index,
                                                &newPathLen)) != NULL) {
-#if defined(linux)
+         char *newRelPath;
+
+#if defined(__linux__)
          if (DnD_UriIsNonFileSchemes(newPath)) {
             /* Try to get local file path for non file uri. */
             GFile *file = g_file_new_for_uri(newPath);
@@ -1424,7 +1495,7 @@ DnDUIX11::SetCPClipboardFromGtk(const Gtk::SelectionData& sd) // IN
          g_debug("%s: Adding newPath '%s' newRelPath '%s'\n", __FUNCTION__,
                newPath, newRelPath);
          fileList.AddFile(newPath, newRelPath);
-#if defined(linux)
+#if defined(__linux__)
          char *newUri = HgfsUri_ConvertFromPathToHgfsUri(newPath, false);
          fileList.AddFileUri(newUri);
          free(newUri);
@@ -1439,7 +1510,7 @@ DnDUIX11::SetCPClipboardFromGtk(const Gtk::SelectionData& sd) // IN
                               DynBuf_GetSize(&buf));
       }
       DynBuf_Destroy(&buf);
-#if defined(linux)
+#if defined(__linux__)
       if (fileList.ToUriClipboard(&buf)) {
          CPClipboard_SetItem(&mClipboard, CPFORMAT_FILELIST_URI, DynBuf_Get(&buf),
                              DynBuf_GetSize(&buf));
@@ -1457,7 +1528,7 @@ DnDUIX11::SetCPClipboardFromGtk(const Gtk::SelectionData& sd) // IN
           && source.size() < DNDMSG_MAX_ARGSZ
           && CPClipboard_SetItem(&mClipboard, CPFORMAT_TEXT, source.c_str(),
                                  source.size() + 1)) {
-         g_debug("%s: Got text, size %"FMTSZ"u\n", __FUNCTION__, source.size());
+         g_debug("%s: Got text, size %" FMTSZ "u\n", __FUNCTION__, source.size());
       } else {
          g_debug("%s: Failed to get text\n", __FUNCTION__);
          return false;
@@ -1473,7 +1544,7 @@ DnDUIX11::SetCPClipboardFromGtk(const Gtk::SelectionData& sd) // IN
           && source.size() < DNDMSG_MAX_ARGSZ
           && CPClipboard_SetItem(&mClipboard, CPFORMAT_RTF, source.c_str(),
                                  source.size() + 1)) {
-         g_debug("%s: Got RTF, size %"FMTSZ"u\n", __FUNCTION__, source.size());
+         g_debug("%s: Got RTF, size %" FMTSZ "u\n", __FUNCTION__, source.size());
          return true;
       } else {
          g_debug("%s: Failed to get text\n", __FUNCTION__ );
@@ -1509,20 +1580,22 @@ DnDUIX11::RequestData(
    guint time)                               // IN: event timestamp
 {
    Glib::RefPtr<Gtk::TargetList> targets;
-   targets = Gtk::TargetList::create(std::list<Gtk::TargetEntry>());
+   targets = Gtk::TargetList::create(std::vector<Gtk::TargetEntry>());
 
    CPClipboard_Clear(&mClipboard);
    mNumPendingRequest = 0;
 
+   Glib::ustring noneType = Gdk::AtomString::to_cpp_type(GDK_NONE);
+
    /*
     * First check file list. If file list is available, all other formats will
     * be ignored.
     */
    targets->add(Glib::ustring(DRAG_TARGET_NAME_URI_LIST));
-   Glib::ustring target = mDetWnd->drag_dest_find_target(dc, targets);
+   Glib::ustring target = mDetWnd->GetWnd()->drag_dest_find_target(dc, targets);
    targets->remove(Glib::ustring(DRAG_TARGET_NAME_URI_LIST));
-   if (target != "") {
-      mDetWnd->drag_get_data(dc, target, time);
+   if (target != noneType) {
+      mDetWnd->GetWnd()->drag_get_data(dc, target, time);
       mNumPendingRequest++;
       return true;
    }
@@ -1532,24 +1605,26 @@ DnDUIX11::RequestData(
    targets->add(Glib::ustring(TARGET_NAME_STRING));
    targets->add(Glib::ustring(TARGET_NAME_TEXT_PLAIN));
    targets->add(Glib::ustring(TARGET_NAME_COMPOUND_TEXT));
-   target = mDetWnd->drag_dest_find_target(dc, targets);
+   target = mDetWnd->GetWnd()->drag_dest_find_target(dc, targets);
    targets->remove(Glib::ustring(TARGET_NAME_STRING));
    targets->remove(Glib::ustring(TARGET_NAME_TEXT_PLAIN));
    targets->remove(Glib::ustring(TARGET_NAME_UTF8_STRING));
    targets->remove(Glib::ustring(TARGET_NAME_COMPOUND_TEXT));
-   if (target != "") {
-      mDetWnd->drag_get_data(dc, target, time);
+   if (target != noneType) {
+      mDetWnd->GetWnd()->drag_get_data(dc, target, time);
       mNumPendingRequest++;
    }
 
    /* Then check RTF. */
    targets->add(Glib::ustring(TARGET_NAME_APPLICATION_RTF));
    targets->add(Glib::ustring(TARGET_NAME_TEXT_RICHTEXT));
-   target = mDetWnd->drag_dest_find_target(dc, targets);
+   targets->add(Glib::ustring(TARGET_NAME_TEXT_RTF));
+   target = mDetWnd->GetWnd()->drag_dest_find_target(dc, targets);
    targets->remove(Glib::ustring(TARGET_NAME_APPLICATION_RTF));
    targets->remove(Glib::ustring(TARGET_NAME_TEXT_RICHTEXT));
-   if (target != "") {
-      mDetWnd->drag_get_data(dc, target, time);
+   targets->remove(Glib::ustring(TARGET_NAME_TEXT_RTF));
+   if (target != noneType) {
+      mDetWnd->GetWnd()->drag_get_data(dc, target, time);
       mNumPendingRequest++;
    }
    return (mNumPendingRequest > 0);
@@ -1567,7 +1642,7 @@ DnDUIX11::RequestData(
  *      name, intended to isolate an individual DND operation's staging directory
  *      name.
  *
- *         E.g. /tmp/VMwareDnD/abcd137/foo → abcd137
+ *         E.g. /tmp/VMwareDnD/abcd137 → abcd137
  *
  * Results:
  *      Returns session directory name on success, empty string otherwise.
@@ -1581,26 +1656,22 @@ DnDUIX11::RequestData(
 std::string
 DnDUIX11::GetLastDirName(const std::string &str)
 {
-   std::string ret;
-   size_t start;
-   size_t end;
-
-   end = str.size() - 1;
-   if (end >= 0 && DIRSEPC == str[end]) {
-      end--;
+   char *baseName;
+   std::string stripSlash = str;
+   char *path = File_StripSlashes(stripSlash.c_str());
+   if (path) {
+      stripSlash = path;
+      free(path);
    }
 
-   if (end <= 0 || str[0] != DIRSEPC) {
-      return "";
+   File_GetPathName(stripSlash.c_str(), NULL, &baseName);
+   if (baseName) {
+      std::string s(baseName);
+      free(baseName);
+      return s;
+   } else {
+      return std::string();
    }
-
-   start = end;
-
-   while (str[start] != DIRSEPC) {
-      start--;
-   }
-
-   return str.substr(start + 1, end - start);
 }
 
 
@@ -1757,9 +1828,13 @@ DnDUIX11::SendFakeXEvents(
       g_debug("%s: unable to get widget\n", __FUNCTION__);
       return false;
    }
-
+#ifndef GTK3
    dndXDisplay = GDK_WINDOW_XDISPLAY(widget->window);
    dndXWindow = GDK_WINDOW_XWINDOW(widget->window);
+#else
+   dndXDisplay = GDK_WINDOW_XDISPLAY(gtk_widget_get_window(widget));
+   dndXWindow = GDK_WINDOW_XID(gtk_widget_get_window(widget));
+#endif
    rootWnd = RootWindow(dndXDisplay, DefaultScreen(dndXDisplay));
 
    /*
@@ -1773,7 +1848,11 @@ DnDUIX11::SendFakeXEvents(
    if (showWidget) {
       g_debug("%s: showing Gtk widget\n", __FUNCTION__);
       gtk_widget_show(widget);
+#ifndef GTK3
       gdk_window_show(widget->window);
+#else
+      gdk_window_show(gtk_widget_get_window(widget));
+#endif
    }
 
    /* Get the current location of the mouse if coordinates weren't provided. */
@@ -1838,24 +1917,47 @@ DnDUIX11::SendFakeXEvents(
        * is okay since the window is invisible and hidden on cancels and DnD
        * finish.
        */
-      XMoveResizeWindow(dndXDisplay, dndXWindow, x - 5 , y - 5, 25, 25);
+      XMoveResizeWindow(dndXDisplay,
+                        dndXWindow,
+                        x - DRAG_DET_WINDOW_WIDTH / 2 ,
+                        y - DRAG_DET_WINDOW_WIDTH / 2,
+                        DRAG_DET_WINDOW_WIDTH,
+                        DRAG_DET_WINDOW_WIDTH);
       XRaiseWindow(dndXDisplay, dndXWindow);
-      g_debug("%s: move wnd to (%d, %d, %d, %d)\n", __FUNCTION__, x - 5, y - 5 , x + 25, y + 25);
+      g_debug("%s: move wnd to (%d, %d, %d, %d)\n",
+              __FUNCTION__,
+              x - DRAG_DET_WINDOW_WIDTH / 2 ,
+              y - DRAG_DET_WINDOW_WIDTH / 2,
+              DRAG_DET_WINDOW_WIDTH,
+              DRAG_DET_WINDOW_WIDTH);
    }
 
    /*
     * Generate mouse movements over the window.  The second one makes ungrabs
     * happen more reliably on KDE, but isn't necessary on GNOME.
     */
-   XTestFakeMotionEvent(dndXDisplay, -1, x, y, CurrentTime);
-   XTestFakeMotionEvent(dndXDisplay, -1, x + 1, y + 1, CurrentTime);
+   if (mUseUInput) {
+#ifdef USE_UINPUT
+      FakeMouse_Move(x, y);
+      FakeMouse_Move(x + 1, y + 1);
+#endif
+   } else {
+      XTestFakeMotionEvent(dndXDisplay, -1, x, y, CurrentTime);
+      XTestFakeMotionEvent(dndXDisplay, -1, x + 1, y + 1, CurrentTime);
+   }
    g_debug("%s: move mouse to (%d, %d) and (%d, %d)\n", __FUNCTION__, x, y, x + 1, y + 1);
 
    if (buttonEvent) {
       g_debug("%s: faking left mouse button %s\n", __FUNCTION__,
               buttonPress ? "press" : "release");
-      XTestFakeButtonEvent(dndXDisplay, 1, buttonPress, CurrentTime);
-      XSync(dndXDisplay, False);
+      if (mUseUInput) {
+#ifdef USE_UINPUT
+         FakeMouse_Click(buttonPress);
+#endif
+      } else {
+         XTestFakeButtonEvent(dndXDisplay, 1, buttonPress, CurrentTime);
+         XSync(dndXDisplay, False);
+      }
 
       if (!buttonPress) {
          /*
@@ -1937,8 +2039,11 @@ DnDUIX11::TryXTestFakeDeviceButtonEvent()
       g_debug("%s: unable to get widget\n", __FUNCTION__);
       return false;
    }
-
+#ifndef GTK3
    dndXDisplay = GDK_WINDOW_XDISPLAY(widget->window);
+#else
+   dndXDisplay = GDK_WINDOW_XDISPLAY(gtk_widget_get_window(widget));
+#endif
 
    /* First get list of all input device. */
    if (!(list = XListInputDevices (dndXDisplay, &numDevices))) {
@@ -2009,15 +2114,14 @@ DnDUIX11::TryXTestFakeDeviceButtonEvent()
 GtkWidget *
 DnDUIX11::GetDetWndAsWidget()
 {
-   GtkInvisible *window;
    GtkWidget *widget = NULL;
 
    if (!mDetWnd) {
       return NULL;
    }
-   window = mDetWnd->gobj();
-   if (window) {
-      widget = GTK_WIDGET(window);
+
+   if (mDetWnd->GetWnd()->gobj()) {
+      widget = GTK_WIDGET(mDetWnd->GetWnd()->gobj());
    }
    return widget;
 }
@@ -2229,7 +2333,7 @@ DnDUIX11::WriteFileContentsToStagingDir()
 
       if (!fileItem[i].cpName.cpName_val ||
           0 == fileItem[i].cpName.cpName_len) {
-         g_debug("%s: invalid fileItem[%"FMTSZ"u].cpName.\n", __FUNCTION__, i);
+         g_debug("%s: invalid fileItem[%" FMTSZ "u].cpName.\n", __FUNCTION__, i);
          goto exit;
       }
 
